@@ -17,8 +17,10 @@
 
 NAMESPACE_BEGIN(CryptoPP)
 
-//! .
-class DL_GroupParameters_IntegerBased : public DL_GroupParameters<Integer>, public ASN1CryptoMaterial
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_GroupParameters<Integer>;
+
+//! _
+class CRYPTOPP_DLL CRYPTOPP_NO_VTABLE DL_GroupParameters_IntegerBased : public DL_GroupParameters<Integer>, public ASN1CryptoMaterial
 {
 	typedef DL_GroupParameters_IntegerBased ThisClass;
 	
@@ -55,6 +57,7 @@ public:
 	Integer ConvertElementToInteger(const Element &element) const
 		{return element;}
 	Integer GetMaxExponent() const;
+	static std::string StaticAlgorithmNamePrefix() {return "";}
 
 	OID GetAlgorithmID() const;
 
@@ -76,9 +79,9 @@ private:
 	Integer m_q;
 };
 
-//! .
+//! _
 template <class GROUP_PRECOMP, class BASE_PRECOMP = DL_FixedBasePrecomputationImpl<CPP_TYPENAME GROUP_PRECOMP::Element> >
-class DL_GroupParameters_IntegerBasedImpl : public DL_GroupParametersImpl<GROUP_PRECOMP, BASE_PRECOMP, DL_GroupParameters_IntegerBased>
+class CRYPTOPP_NO_VTABLE DL_GroupParameters_IntegerBasedImpl : public DL_GroupParametersImpl<GROUP_PRECOMP, BASE_PRECOMP, DL_GroupParameters_IntegerBased>
 {
 	typedef DL_GroupParameters_IntegerBasedImpl<GROUP_PRECOMP, BASE_PRECOMP> ThisClass;
 
@@ -93,25 +96,27 @@ public:
 		{AssignFromHelper<DL_GroupParameters_IntegerBased>(this, source);}
 
 	// DL_GroupParameters
-	const DL_FixedBasePrecomputation<Element> & GetBasePrecomputation() const {return m_gpc;}
-	DL_FixedBasePrecomputation<Element> & AccessBasePrecomputation() {return m_gpc;}
+	const DL_FixedBasePrecomputation<Element> & GetBasePrecomputation() const {return this->m_gpc;}
+	DL_FixedBasePrecomputation<Element> & AccessBasePrecomputation() {return this->m_gpc;}
 
 	// IntegerGroupParameters
-	const Integer & GetModulus() const {return m_groupPrecomputation.GetModulus();}
-	const Integer & GetGenerator() const {return m_gpc.GetBase(GetGroupPrecomputation());}
+	const Integer & GetModulus() const {return this->m_groupPrecomputation.GetModulus();}
+    const Integer & GetGenerator() const {return this->m_gpc.GetBase(this->GetGroupPrecomputation());}
 
 	void SetModulusAndSubgroupGenerator(const Integer &p, const Integer &g)		// these have to be set together
-		{m_groupPrecomputation.SetModulus(p); m_gpc.SetBase(GetGroupPrecomputation(), g); ParametersChanged();}
+		{this->m_groupPrecomputation.SetModulus(p); this->m_gpc.SetBase(this->GetGroupPrecomputation(), g); this->ParametersChanged();}
 
 	// non-inherited
 	bool operator==(const DL_GroupParameters_IntegerBasedImpl<GROUP_PRECOMP, BASE_PRECOMP> &rhs) const
-		{return GetModulus() == rhs.GetModulus() && GetGenerator() == rhs.GetGenerator() && GetSubgroupOrder() == rhs.GetSubgroupOrder();}
+		{return GetModulus() == rhs.GetModulus() && GetGenerator() == rhs.GetGenerator() && this->GetSubgroupOrder() == rhs.GetSubgroupOrder();}
 	bool operator!=(const DL_GroupParameters_IntegerBasedImpl<GROUP_PRECOMP, BASE_PRECOMP> &rhs) const
 		{return !operator==(rhs);}
 };
 
-//! .
-class DL_GroupParameters_GFP : public DL_GroupParameters_IntegerBasedImpl<ModExpPrecomputation>
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_GroupParameters_IntegerBasedImpl<ModExpPrecomputation>;
+
+//! GF(p) group parameters
+class CRYPTOPP_DLL DL_GroupParameters_GFP : public DL_GroupParameters_IntegerBasedImpl<ModExpPrecomputation>
 {
 public:
 	// DL_GroupParameters
@@ -132,8 +137,8 @@ protected:
 	int GetFieldType() const {return 1;}
 };
 
-//! .
-class DL_GroupParameters_GFP_DefaultSafePrime : public DL_GroupParameters_GFP
+//! GF(p) group parameters that default to same primes
+class CRYPTOPP_DLL DL_GroupParameters_GFP_DefaultSafePrime : public DL_GroupParameters_GFP
 {
 public:
 	typedef NoCofactorMultiplication DefaultCofactorOption;
@@ -142,7 +147,7 @@ protected:
 	unsigned int GetDefaultSubgroupOrderSize(unsigned int modulusSize) const {return modulusSize-1;}
 };
 
-//! .
+//! GDSA algorithm
 template <class T>
 class DL_Algorithm_GDSA : public DL_ElgamalLikeSignatureAlgorithm<T>
 {
@@ -172,18 +177,15 @@ public:
 	}
 };
 
-//! .
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_Algorithm_GDSA<Integer>;
+
+//! NR algorithm
 template <class T>
 class DL_Algorithm_NR : public DL_ElgamalLikeSignatureAlgorithm<T>
 {
 public:
 	static const char * StaticAlgorithmName() {return "NR";}
 
-	Integer EncodeDigest(unsigned int modulusBits, const byte *digest, unsigned int digestLen) const
-	{
-		return NR_EncodeDigest(modulusBits, digest, digestLen);
-	}
-
 	void Sign(const DL_GroupParameters<T> &params, const Integer &x, const Integer &k, const Integer &e, Integer &r, Integer &s) const
 	{
 		const Integer &q = params.GetSubgroupOrder();
@@ -210,39 +212,39 @@ class DL_PublicKey_GFP : public DL_PublicKeyImpl<GP>
 {
 public:
 	void Initialize(const DL_GroupParameters_IntegerBased &params, const Integer &y)
-		{AccessGroupParameters().Initialize(params); SetPublicElement(y);}
+		{this->AccessGroupParameters().Initialize(params); this->SetPublicElement(y);}
 	void Initialize(const Integer &p, const Integer &g, const Integer &y)
-		{AccessGroupParameters().Initialize(p, g); SetPublicElement(y);}
+		{this->AccessGroupParameters().Initialize(p, g); this->SetPublicElement(y);}
 	void Initialize(const Integer &p, const Integer &q, const Integer &g, const Integer &y)
-		{AccessGroupParameters().Initialize(p, q, g); SetPublicElement(y);}
+		{this->AccessGroupParameters().Initialize(p, q, g); this->SetPublicElement(y);}
 
 	// X509PublicKey
 	void BERDecodeKey(BufferedTransformation &bt)
-		{SetPublicElement(Integer(bt));}
+		{this->SetPublicElement(Integer(bt));}
 	void DEREncodeKey(BufferedTransformation &bt) const
-		{GetPublicElement().DEREncode(bt);}
+		{this->GetPublicElement().DEREncode(bt);}
 };
 
-//! .
+//! DL private key (in GF(p) groups)
 template <class GP>
 class DL_PrivateKey_GFP : public DL_PrivateKeyImpl<GP>
 {
 public:
 	void Initialize(RandomNumberGenerator &rng, unsigned int modulusBits)
-		{GenerateRandomWithKeySize(rng, modulusBits);}
+		{this->GenerateRandomWithKeySize(rng, modulusBits);}
 	void Initialize(RandomNumberGenerator &rng, const Integer &p, const Integer &g)
-		{GenerateRandom(rng, MakeParameters("Modulus", p)("SubgroupGenerator", g));}
+		{this->GenerateRandom(rng, MakeParameters("Modulus", p)("SubgroupGenerator", g));}
 	void Initialize(RandomNumberGenerator &rng, const Integer &p, const Integer &q, const Integer &g)
-		{GenerateRandom(rng, MakeParameters("Modulus", p)("SubgroupOrder", q)("SubgroupGenerator", g));}
+		{this->GenerateRandom(rng, MakeParameters("Modulus", p)("SubgroupOrder", q)("SubgroupGenerator", g));}
 	void Initialize(const DL_GroupParameters_IntegerBased &params, const Integer &x)
-		{AccessGroupParameters().Initialize(params); SetPrivateExponent(x);}
+		{this->AccessGroupParameters().Initialize(params); this->SetPrivateExponent(x);}
 	void Initialize(const Integer &p, const Integer &g, const Integer &x)
-		{AccessGroupParameters().Initialize(p, g); SetPrivateExponent(x);}
+		{this->AccessGroupParameters().Initialize(p, g); this->SetPrivateExponent(x);}
 	void Initialize(const Integer &p, const Integer &q, const Integer &g, const Integer &x)
-		{AccessGroupParameters().Initialize(p, q, g); SetPrivateExponent(x);}
+		{this->AccessGroupParameters().Initialize(p, q, g); this->SetPrivateExponent(x);}
 };
 
-//! .
+//! DL signing/verification keys (in GF(p) groups)
 struct DL_SignatureKeys_GFP
 {
 	typedef DL_GroupParameters_GFP GroupParameters;
@@ -250,7 +252,7 @@ struct DL_SignatureKeys_GFP
 	typedef DL_PrivateKey_GFP<GroupParameters> PrivateKey;
 };
 
-//! .
+//! DL encryption/decryption keys (in GF(p) groups)
 struct DL_CryptoKeys_GFP
 {
 	typedef DL_GroupParameters_GFP_DefaultSafePrime GroupParameters;
@@ -272,14 +274,14 @@ public:
 
 			if (seq.EndReached())
 			{
-				AccessGroupParameters().Initialize(v1, v1/2, v2);
-				SetPublicElement(v3);
+				this->AccessGroupParameters().Initialize(v1, v1/2, v2);
+				this->SetPublicElement(v3);
 			}
 			else
 			{
 				Integer v4(seq);
-				AccessGroupParameters().Initialize(v1, v2, v3);
-				SetPublicElement(v4);
+				this->AccessGroupParameters().Initialize(v1, v2, v3);
+				this->SetPublicElement(v4);
 			}
 
 		seq.MessageEnd();
@@ -288,11 +290,11 @@ public:
 	void DEREncode(BufferedTransformation &bt) const
 	{
 		DERSequenceEncoder seq(bt);
-			GetGroupParameters().GetModulus().DEREncode(seq);
-			if (GetGroupParameters().GetCofactor() != 2)
-				GetGroupParameters().GetSubgroupOrder().DEREncode(seq);
-			GetGroupParameters().GetGenerator().DEREncode(seq);
-			GetPublicElement().DEREncode(seq);
+			this->GetGroupParameters().GetModulus().DEREncode(seq);
+			if (this->GetGroupParameters().GetCofactor() != 2)
+				this->GetGroupParameters().GetSubgroupOrder().DEREncode(seq);
+			this->GetGroupParameters().GetGenerator().DEREncode(seq);
+			this->GetPublicElement().DEREncode(seq);
 		seq.MessageEnd();
 	}
 };
@@ -312,14 +314,14 @@ public:
 
 			if (seq.EndReached())
 			{
-				AccessGroupParameters().Initialize(v1, v1/2, v2);
-				SetPrivateExponent(v4 % (v1/2));	// some old keys may have x >= q
+				this->AccessGroupParameters().Initialize(v1, v1/2, v2);
+				this->SetPrivateExponent(v4 % (v1/2));	// some old keys may have x >= q
 			}
 			else
 			{
 				Integer v5(seq);
-				AccessGroupParameters().Initialize(v1, v2, v3);
-				SetPrivateExponent(v5);
+				this->AccessGroupParameters().Initialize(v1, v2, v3);
+				this->SetPrivateExponent(v5);
 			}
 
 		seq.MessageEnd();
@@ -328,12 +330,12 @@ public:
 	void DEREncode(BufferedTransformation &bt) const
 	{
 		DERSequenceEncoder seq(bt);
-			GetGroupParameters().GetModulus().DEREncode(seq);
-			if (GetGroupParameters().GetCofactor() != 2)
-				GetGroupParameters().GetSubgroupOrder().DEREncode(seq);
-			GetGroupParameters().GetGenerator().DEREncode(seq);
-			GetGroupParameters().ExponentiateBase(GetPrivateExponent()).DEREncode(seq);
-			GetPrivateExponent().DEREncode(seq);
+			this->GetGroupParameters().GetModulus().DEREncode(seq);
+			if (this->GetGroupParameters().GetCofactor() != 2)
+				this->GetGroupParameters().GetSubgroupOrder().DEREncode(seq);
+			this->GetGroupParameters().GetGenerator().DEREncode(seq);
+			this->GetGroupParameters().ExponentiateBase(this->GetPrivateExponent()).DEREncode(seq);
+			this->GetPrivateExponent().DEREncode(seq);
 		seq.MessageEnd();
 	}
 };
@@ -358,20 +360,20 @@ struct NR : public DL_SS<
 {
 };
 
-//! .
-class DL_GroupParameters_DSA : public DL_GroupParameters_GFP
+//! DSA group parameters, these are GF(p) group parameters that are allowed by the DSA standard
+class CRYPTOPP_DLL DL_GroupParameters_DSA : public DL_GroupParameters_GFP
 {
 public:
 	/*! also checks that the lengths of p and q are allowed by the DSA standard */
 	bool ValidateGroup(RandomNumberGenerator &rng, unsigned int level) const;
 	/*! parameters: (ModulusSize), or (Modulus, SubgroupOrder, SubgroupGenerator) */
-	/*! ModulusSize must be between 512 and 1024, and divisible by 64 */
+	/*! ModulusSize must be between DSA::MIN_PRIME_LENGTH and DSA::MAX_PRIME_LENGTH, and divisible by DSA::PRIME_LENGTH_MULTIPLE */
 	void GenerateRandom(RandomNumberGenerator &rng, const NameValuePairs &alg);
 };
 
 struct DSA;
 
-//! .
+//! DSA keys
 struct DL_Keys_DSA
 {
 	typedef DL_PublicKey_GFP<DL_GroupParameters_DSA> PublicKey;
@@ -379,7 +381,7 @@ struct DL_Keys_DSA
 };
 
 //! <a href="http://www.weidai.com/scan-mirror/sig.html#DSA">DSA</a>
-struct DSA : public DL_SS<
+struct CRYPTOPP_DLL DSA : public DL_SS<
 	DL_Keys_DSA, 
 	DL_Algorithm_GDSA<Integer>, 
 	DL_SignatureMessageEncodingMethod_DSA,
@@ -399,6 +401,7 @@ struct DSA : public DL_SS<
 	static bool IsValidPrimeLength(unsigned int pbits)
 		{return pbits >= MIN_PRIME_LENGTH && pbits <= MAX_PRIME_LENGTH && pbits % PRIME_LENGTH_MULTIPLE == 0;}
 
+	//! FIPS 186-2 Change Notice 1 changed the minimum modulus length to 1024
 	enum {
 #if (DSA_1024_BIT_MODULUS_ONLY)
 		MIN_PRIME_LENGTH = 1024,
@@ -408,18 +411,23 @@ struct DSA : public DL_SS<
 		MAX_PRIME_LENGTH = 1024, PRIME_LENGTH_MULTIPLE = 64};
 };
 
-//! .
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_PublicKey_GFP<DL_GroupParameters_DSA>;
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_PrivateKey_GFP<DL_GroupParameters_DSA>;
+CRYPTOPP_DLL_TEMPLATE_CLASS DL_PrivateKey_WithSignaturePairwiseConsistencyTest<DL_PrivateKey_GFP<DL_GroupParameters_DSA>, DSA>;
+
+//! the XOR encryption method, for use with DL-based cryptosystems
 template <class MAC, bool DHAES_MODE>
 class DL_EncryptionAlgorithm_Xor : public DL_SymmetricEncryptionAlgorithm
 {
 public:
-	unsigned int GetSymmetricKeyLength(unsigned int plainTextLength) const
-		{return plainTextLength + MAC::DEFAULT_KEYLENGTH;}
-	unsigned int GetSymmetricCiphertextLength(unsigned int plainTextLength) const
-		{return plainTextLength + MAC::DIGESTSIZE;}
-	unsigned int GetMaxSymmetricPlaintextLength(unsigned int cipherTextLength) const
-		{return SaturatingSubtract(cipherTextLength, (unsigned int)MAC::DIGESTSIZE);}
-	void SymmetricEncrypt(RandomNumberGenerator &rng, const byte *key, const byte *plainText, unsigned int plainTextLength, byte *cipherText) const
+	bool ParameterSupported(const char *name) const {return strcmp(name, Name::EncodingParameters()) == 0;}
+	unsigned int GetSymmetricKeyLength(unsigned int plaintextLength) const
+		{return plaintextLength + MAC::DEFAULT_KEYLENGTH;}
+	unsigned int GetSymmetricCiphertextLength(unsigned int plaintextLength) const
+		{return plaintextLength + MAC::DIGESTSIZE;}
+	unsigned int GetMaxSymmetricPlaintextLength(unsigned int ciphertextLength) const
+		{return SaturatingSubtract(ciphertextLength, (unsigned int)MAC::DIGESTSIZE);}
+	void SymmetricEncrypt(RandomNumberGenerator &rng, const byte *key, const byte *plaintext, unsigned int plaintextLength, byte *ciphertext, const NameValuePairs &parameters) const
 	{
 		const byte *cipherKey, *macKey;
 		if (DHAES_MODE)
@@ -430,22 +438,27 @@ public:
 		else
 		{
 			cipherKey = key;
-			macKey = key + plainTextLength;
+			macKey = key + plaintextLength;
 		}
 
-		xorbuf(cipherText, plainText, cipherKey, plainTextLength);
+		ConstByteArrayParameter encodingParameters;
+		parameters.GetValue(Name::EncodingParameters(), encodingParameters);
+
+		xorbuf(ciphertext, plaintext, cipherKey, plaintextLength);
 		MAC mac(macKey);
-		mac.Update(cipherText, plainTextLength);
+		mac.Update(ciphertext, plaintextLength);
+		mac.Update(encodingParameters.begin(), encodingParameters.size());
 		if (DHAES_MODE)
 		{
-			const byte L[8] = {0,0,0,0,0,0,0,0};
+			byte L[8] = {0,0,0,0};
+			UnalignedPutWord(BIG_ENDIAN_ORDER, L+4, word32(encodingParameters.size()));
 			mac.Update(L, 8);
 		}
-		mac.Final(cipherText + plainTextLength);
+		mac.Final(ciphertext + plaintextLength);
 	}
-	DecodingResult SymmetricDecrypt(const byte *key, const byte *cipherText, unsigned int cipherTextLength, byte *plainText) const
+	DecodingResult SymmetricDecrypt(const byte *key, const byte *ciphertext, unsigned int ciphertextLength, byte *plaintext, const NameValuePairs &parameters) const
 	{
-		unsigned int plainTextLength = GetMaxSymmetricPlaintextLength(cipherTextLength);
+		unsigned int plaintextLength = GetMaxSymmetricPlaintextLength(ciphertextLength);
 		const byte *cipherKey, *macKey;
 		if (DHAES_MODE)
 		{
@@ -455,30 +468,36 @@ public:
 		else
 		{
 			cipherKey = key;
-			macKey = key + plainTextLength;
+			macKey = key + plaintextLength;
 		}
 
+		ConstByteArrayParameter encodingParameters;
+		parameters.GetValue(Name::EncodingParameters(), encodingParameters);
+
 		MAC mac(macKey);
-		mac.Update(cipherText, plainTextLength);
+		mac.Update(ciphertext, plaintextLength);
+		mac.Update(encodingParameters.begin(), encodingParameters.size());
 		if (DHAES_MODE)
 		{
-			const byte L[8] = {0,0,0,0,0,0,0,0};
+			byte L[8] = {0,0,0,0};
+			UnalignedPutWord(BIG_ENDIAN_ORDER, L+4, word32(encodingParameters.size()));
 			mac.Update(L, 8);
 		}
-		if (!mac.Verify(cipherText + plainTextLength))
+		if (!mac.Verify(ciphertext + plaintextLength))
 			return DecodingResult();
 
-		xorbuf(plainText, cipherText, cipherKey, plainTextLength);
-		return DecodingResult(plainTextLength);
+		xorbuf(plaintext, ciphertext, cipherKey, plaintextLength);
+		return DecodingResult(plaintextLength);
 	}
 };
 
-//! .
+//! _
 template <class T, bool DHAES_MODE, class KDF>
 class DL_KeyDerivationAlgorithm_P1363 : public DL_KeyDerivationAlgorithm<T>
 {
 public:
-	void Derive(const DL_GroupParameters<T> &params, byte *derivedKey, unsigned int derivedLength, const T &agreedElement, const T &ephemeralPublicKey) const
+	bool ParameterSupported(const char *name) const {return strcmp(name, Name::KeyDerivationParameters()) == 0;}
+	void Derive(const DL_GroupParameters<T> &params, byte *derivedKey, unsigned int derivedLength, const T &agreedElement, const T &ephemeralPublicKey, const NameValuePairs &parameters) const
 	{
 		SecByteBlock agreedSecret;
 		if (DHAES_MODE)
@@ -493,7 +512,9 @@ public:
 			params.EncodeElement(false, agreedElement, agreedSecret);
 		}
 
-		KDF::DeriveKey(derivedKey, derivedLength, agreedSecret, agreedSecret.size());
+		ConstByteArrayParameter derivationParameters;
+		parameters.GetValue(Name::KeyDerivationParameters(), derivationParameters);
+		KDF::DeriveKey(derivedKey, derivedLength, agreedSecret, agreedSecret.size(), derivationParameters.begin(), derivationParameters.size());
 	}
 };
 
